# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import gcs_service_pb2 as src_dot_ray_dot_protobuf_dot_gcs__service__pb2


class JobInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddJob = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/AddJob',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobReply.FromString,
        )
    self.MarkJobFinished = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/MarkJobFinished',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedReply.FromString,
        )
    self.GetAllJobInfo = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/GetAllJobInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoReply.FromString,
        )
    self.ReportJobError = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/ReportJobError',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorReply.FromString,
        )
    self.GetNextJobID = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/GetNextJobID',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDReply.FromString,
        )


class JobInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddJob(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def MarkJobFinished(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllJobInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ReportJobError(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNextJobID(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_JobInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddJob': grpc.unary_unary_rpc_method_handler(
          servicer.AddJob,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobReply.SerializeToString,
      ),
      'MarkJobFinished': grpc.unary_unary_rpc_method_handler(
          servicer.MarkJobFinished,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedReply.SerializeToString,
      ),
      'GetAllJobInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllJobInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoReply.SerializeToString,
      ),
      'ReportJobError': grpc.unary_unary_rpc_method_handler(
          servicer.ReportJobError,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorReply.SerializeToString,
      ),
      'GetNextJobID': grpc.unary_unary_rpc_method_handler(
          servicer.GetNextJobID,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.JobInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class ActorInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RegisterActor = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/RegisterActor',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorReply.FromString,
        )
    self.CreateActor = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/CreateActor',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorReply.FromString,
        )
    self.GetActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoReply.FromString,
        )
    self.GetNamedActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetNamedActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoReply.FromString,
        )
    self.GetAllActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetAllActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoReply.FromString,
        )
    self.KillActorViaGcs = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/KillActorViaGcs',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsReply.FromString,
        )


class ActorInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RegisterActor(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateActor(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNamedActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def KillActorViaGcs(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ActorInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RegisterActor': grpc.unary_unary_rpc_method_handler(
          servicer.RegisterActor,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorReply.SerializeToString,
      ),
      'CreateActor': grpc.unary_unary_rpc_method_handler(
          servicer.CreateActor,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorReply.SerializeToString,
      ),
      'GetActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoReply.SerializeToString,
      ),
      'GetNamedActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNamedActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoReply.SerializeToString,
      ),
      'GetAllActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoReply.SerializeToString,
      ),
      'KillActorViaGcs': grpc.unary_unary_rpc_method_handler(
          servicer.KillActorViaGcs,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.ActorInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class NodeInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RegisterNode = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/RegisterNode',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeReply.FromString,
        )
    self.UnregisterNode = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/UnregisterNode',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeReply.FromString,
        )
    self.GetAllNodeInfo = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/GetAllNodeInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoReply.FromString,
        )
    self.GetInternalConfig = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/GetInternalConfig',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigReply.FromString,
        )


class NodeInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RegisterNode(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnregisterNode(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllNodeInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetInternalConfig(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_NodeInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RegisterNode': grpc.unary_unary_rpc_method_handler(
          servicer.RegisterNode,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeReply.SerializeToString,
      ),
      'UnregisterNode': grpc.unary_unary_rpc_method_handler(
          servicer.UnregisterNode,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeReply.SerializeToString,
      ),
      'GetAllNodeInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllNodeInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoReply.SerializeToString,
      ),
      'GetInternalConfig': grpc.unary_unary_rpc_method_handler(
          servicer.GetInternalConfig,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.NodeInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class HeartbeatInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ReportHeartbeat = channel.unary_unary(
        '/ray.rpc.HeartbeatInfoGcsService/ReportHeartbeat',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatReply.FromString,
        )


class HeartbeatInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ReportHeartbeat(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_HeartbeatInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ReportHeartbeat': grpc.unary_unary_rpc_method_handler(
          servicer.ReportHeartbeat,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.HeartbeatInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class ObjectInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetObjectLocations = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/GetObjectLocations',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsReply.FromString,
        )
    self.GetAllObjectLocations = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/GetAllObjectLocations',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsReply.FromString,
        )
    self.AddObjectLocation = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/AddObjectLocation',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationReply.FromString,
        )
    self.RemoveObjectLocation = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/RemoveObjectLocation',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationReply.FromString,
        )


class ObjectInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetObjectLocations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllObjectLocations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddObjectLocation(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveObjectLocation(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ObjectInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetObjectLocations': grpc.unary_unary_rpc_method_handler(
          servicer.GetObjectLocations,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsReply.SerializeToString,
      ),
      'GetAllObjectLocations': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllObjectLocations,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsReply.SerializeToString,
      ),
      'AddObjectLocation': grpc.unary_unary_rpc_method_handler(
          servicer.AddObjectLocation,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationReply.SerializeToString,
      ),
      'RemoveObjectLocation': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveObjectLocation,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.ObjectInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class TaskInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddTask = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AddTask',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskReply.FromString,
        )
    self.GetTask = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/GetTask',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskReply.FromString,
        )
    self.AddTaskLease = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AddTaskLease',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseReply.FromString,
        )
    self.GetTaskLease = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/GetTaskLease',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseReply.FromString,
        )
    self.AttemptTaskReconstruction = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AttemptTaskReconstruction',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionReply.FromString,
        )


class TaskInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddTaskLease(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTaskLease(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AttemptTaskReconstruction(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TaskInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddTask': grpc.unary_unary_rpc_method_handler(
          servicer.AddTask,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskReply.SerializeToString,
      ),
      'GetTask': grpc.unary_unary_rpc_method_handler(
          servicer.GetTask,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskReply.SerializeToString,
      ),
      'AddTaskLease': grpc.unary_unary_rpc_method_handler(
          servicer.AddTaskLease,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseReply.SerializeToString,
      ),
      'GetTaskLease': grpc.unary_unary_rpc_method_handler(
          servicer.GetTaskLease,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseReply.SerializeToString,
      ),
      'AttemptTaskReconstruction': grpc.unary_unary_rpc_method_handler(
          servicer.AttemptTaskReconstruction,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.TaskInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class StatsGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddProfileData = channel.unary_unary(
        '/ray.rpc.StatsGcsService/AddProfileData',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataReply.FromString,
        )
    self.GetAllProfileInfo = channel.unary_unary(
        '/ray.rpc.StatsGcsService/GetAllProfileInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoReply.FromString,
        )


class StatsGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddProfileData(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllProfileInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_StatsGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddProfileData': grpc.unary_unary_rpc_method_handler(
          servicer.AddProfileData,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataReply.SerializeToString,
      ),
      'GetAllProfileInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllProfileInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.StatsGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class WorkerInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ReportWorkerFailure = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/ReportWorkerFailure',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureReply.FromString,
        )
    self.GetWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/GetWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoReply.FromString,
        )
    self.GetAllWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/GetAllWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoReply.FromString,
        )
    self.AddWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/AddWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoReply.FromString,
        )


class WorkerInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ReportWorkerFailure(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WorkerInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ReportWorkerFailure': grpc.unary_unary_rpc_method_handler(
          servicer.ReportWorkerFailure,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureReply.SerializeToString,
      ),
      'GetWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoReply.SerializeToString,
      ),
      'GetAllWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoReply.SerializeToString,
      ),
      'AddWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.AddWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.WorkerInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class PlacementGroupInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.CreatePlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/CreatePlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupReply.FromString,
        )
    self.RemovePlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/RemovePlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupReply.FromString,
        )
    self.GetPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupReply.FromString,
        )
    self.GetNamedPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetNamedPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupReply.FromString,
        )
    self.GetAllPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetAllPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupReply.FromString,
        )
    self.WaitPlacementGroupUntilReady = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/WaitPlacementGroupUntilReady',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyReply.FromString,
        )


class PlacementGroupInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def CreatePlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemovePlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNamedPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def WaitPlacementGroupUntilReady(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PlacementGroupInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'CreatePlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.CreatePlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupReply.SerializeToString,
      ),
      'RemovePlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.RemovePlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupReply.SerializeToString,
      ),
      'GetPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupReply.SerializeToString,
      ),
      'GetNamedPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetNamedPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupReply.SerializeToString,
      ),
      'GetAllPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupReply.SerializeToString,
      ),
      'WaitPlacementGroupUntilReady': grpc.unary_unary_rpc_method_handler(
          servicer.WaitPlacementGroupUntilReady,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.PlacementGroupInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class InternalKVGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.InternalKVGet = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVGet',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetReply.FromString,
        )
    self.InternalKVPut = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVPut',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutReply.FromString,
        )
    self.InternalKVDel = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVDel',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelReply.FromString,
        )
    self.InternalKVExists = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVExists',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsReply.FromString,
        )
    self.InternalKVKeys = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVKeys',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysReply.FromString,
        )


class InternalKVGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def InternalKVGet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVPut(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVDel(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVExists(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVKeys(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_InternalKVGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'InternalKVGet': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVGet,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetReply.SerializeToString,
      ),
      'InternalKVPut': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVPut,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutReply.SerializeToString,
      ),
      'InternalKVDel': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVDel,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelReply.SerializeToString,
      ),
      'InternalKVExists': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVExists,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsReply.SerializeToString,
      ),
      'InternalKVKeys': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVKeys,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.InternalKVGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class NodeResourceInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesReply.FromString,
        )
    self.UpdateResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/UpdateResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesReply.FromString,
        )
    self.DeleteResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/DeleteResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesReply.FromString,
        )
    self.GetAllAvailableResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetAllAvailableResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesReply.FromString,
        )
    self.ReportResourceUsage = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/ReportResourceUsage',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageReply.FromString,
        )
    self.GetAllResourceUsage = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetAllResourceUsage',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageReply.FromString,
        )


class NodeResourceInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllAvailableResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ReportResourceUsage(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllResourceUsage(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_NodeResourceInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetResources': grpc.unary_unary_rpc_method_handler(
          servicer.GetResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesReply.SerializeToString,
      ),
      'UpdateResources': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesReply.SerializeToString,
      ),
      'DeleteResources': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesReply.SerializeToString,
      ),
      'GetAllAvailableResources': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllAvailableResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesReply.SerializeToString,
      ),
      'ReportResourceUsage': grpc.unary_unary_rpc_method_handler(
          servicer.ReportResourceUsage,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageReply.SerializeToString,
      ),
      'GetAllResourceUsage': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllResourceUsage,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.NodeResourceInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import os
import asyncio
import logging
from typing import Union
from grpc.experimental import aio as aiogrpc

import ray.new_dashboard.utils as dashboard_utils
import ray.new_dashboard.consts as dashboard_consts
from ray.ray_constants import env_bool
from ray.new_dashboard.utils import async_loop_forever, create_task
from ray.new_dashboard.modules.event import event_consts
from ray.new_dashboard.modules.event.event_utils import monitor_events
from ray.core.generated import event_pb2
from ray.core.generated import event_pb2_grpc

logger = logging.getLogger(__name__)
routes = dashboard_utils.ClassMethodRouteTable


@dashboard_utils.dashboard_module(
    enable=env_bool(event_consts.EVENT_MODULE_ENVIRONMENT_KEY, False))
class EventAgent(dashboard_utils.DashboardAgentModule):
    def __init__(self, dashboard_agent):
        super().__init__(dashboard_agent)
        self._event_dir = os.path.join(self._dashboard_agent.log_dir, "events")
        os.makedirs(self._event_dir, exist_ok=True)
        self._monitor: Union[asyncio.Task, None] = None
        self._stub: Union[event_pb2_grpc.ReportEventServiceStub, None] = None
        self._cached_events = asyncio.Queue(
            event_consts.EVENT_AGENT_CACHE_SIZE)
        logger.info("Event agent cache buffer size: %s",
                    self._cached_events.maxsize)

    async def _connect_to_dashboard(self):
        """ Connect to the dashboard. If the dashboard is not started, then
        this method will never returns.

        Returns:
            The ReportEventServiceStub object.
        """
        while True:
            try:
                aioredis = self._dashboard_agent.aioredis_client
                dashboard_rpc_address = await aioredis.get(
                    dashboard_consts.REDIS_KEY_DASHBOARD_RPC)
                if dashboard_rpc_address:
                    logger.info("Report events to %s", dashboard_rpc_address)
                    options = (("grpc.enable_http_proxy", 0), )
                    channel = aiogrpc.insecure_channel(
                        dashboard_rpc_address, options=options)
                    return event_pb2_grpc.ReportEventServiceStub(channel)
            except Exception:
                logger.exception("Connect to dashboard failed.")
            await asyncio.sleep(
                event_consts.RETRY_CONNECT_TO_DASHBOARD_INTERVAL_SECONDS)

    @async_loop_forever(event_consts.EVENT_AGENT_REPORT_INTERVAL_SECONDS)
    async def report_events(self):
        """ Report events from cached events queue. Reconnect to dashboard if
        report failed. Log error after retry EVENT_AGENT_RETRY_TIMES.

        This method will never returns.
        """
        data = await self._cached_events.get()
        for _ in range(event_consts.EVENT_AGENT_RETRY_TIMES):
            try:
                logger.info("Report %s events.", len(data))
                request = event_pb2.ReportEventsRequest(event_strings=data)
                await self._stub.ReportEvents(request)
                break
            except Exception:
                logger.exception("Report event failed, reconnect to the "
                                 "dashboard.")
                self._stub = await self._connect_to_dashboard()
        else:
            data_str = str(data)
            limit = event_consts.LOG_ERROR_EVENT_STRING_LENGTH_LIMIT
            logger.error("Report event failed: %s",
                         data_str[:limit] + (data_str[limit:] and "..."))

    async def run(self, server):
        # Connect to dashboard.
        self._stub = await self._connect_to_dashboard()
        # Start monitor task.
        self._monitor = monitor_events(
            self._event_dir,
            lambda data: create_task(self._cached_events.put(data)),
            source_types=event_consts.EVENT_AGENT_MONITOR_SOURCE_TYPES)
        # Start reporting events.
        await self.report_events()



import argparse
import json
import logging
import os

from ray._private.utils import import_attr
logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(
    description=(
        "Set up the environment for a Ray worker and launch the worker."))

parser.add_argument(
    "--worker-setup-hook",
    type=str,
    help="the module path to a Python function to run to set up the "
    "environment for a worker and launch the worker.")

parser.add_argument(
    "--serialized-runtime-env",
    type=str,
    help="the serialized parsed runtime env dict")


def get_tmp_dir(remaining_args):
    for arg in remaining_args:
        if arg.startswith("--temp-dir="):
            return arg[11:]
    return None


def start_worker_in_container(container_option, args, remaining_args):
    worker_setup_hook = args.worker_setup_hook
    last_period_idx = worker_setup_hook.rfind(".")
    module_name = worker_setup_hook[:last_period_idx]
    # python -m ray.workers.setup_runtime_env --session-dir=
    # default_worker.py --node-ip-address= ...
    entrypoint_args = ["-m"]
    entrypoint_args.append(module_name)
    # replace default_worker.py path
    if container_option.get("worker_path"):
        remaining_args[1] = container_option.get("worker_path")
    entrypoint_args.extend(remaining_args)
    # setup_runtime_env will install conda,pip according to
    # serialized-runtime-env, so add this argument
    entrypoint_args.append("--serialized-runtime-env")
    entrypoint_args.append(args.serialized_runtime_env or "{}")
    # now we will start a container, add argument worker-shim-pid
    entrypoint_args.append("--worker-shim-pid={}".format(os.getpid()))

    tmp_dir = get_tmp_dir(remaining_args)
    if not tmp_dir:
        logger.error(
            "failed to get tmp_dir, the args: {}".format(remaining_args))

    container_driver = "podman"
    # todo add cgroup config
    # todo flag "--rm"
    container_command = [
        container_driver, "run", "-v", tmp_dir + ":" + tmp_dir,
        "--cgroup-manager=cgroupfs", "--network=host", "--pid=host",
        "--ipc=host", "--env-host"
    ]
    container_command.append("--env")
    container_command.append("RAY_RAYLET_PID=" + str(os.getppid()))
    if container_option.get("run_options"):
        container_command.extend(container_option.get("run_options"))

    container_command.append("--entrypoint")
    container_command.append("python")
    container_command.append(container_option.get("image"))
    container_command.extend(entrypoint_args)
    logger.warning("start worker in container: {}".format(container_command))
    os.execvp(container_driver, container_command)


if __name__ == "__main__":
    args, remaining_args = parser.parse_known_args()
    runtime_env: dict = json.loads(args.serialized_runtime_env or "{}")
    container_option = runtime_env.get("container")
    if container_option and container_option.get("image"):
        start_worker_in_container(container_option, args, remaining_args)
    else:
        remaining_args.append("--serialized-runtime-env")
        remaining_args.append(args.serialized_runtime_env or "{}")
        setup = import_attr(args.worker_setup_hook)
        setup(remaining_args)



class BaseWorkerCmd:
    """
    base worker cmd
    """

    def __init__(self, actor_id):
        self.from_actor_id = actor_id


class WorkerCommitReport(BaseWorkerCmd):
    """
    worker commit report
    """

    def __init__(self, actor_id, commit_checkpoint_id):
        super().__init__(actor_id)
        self.commit_checkpoint_id = commit_checkpoint_id


class WorkerRollbackRequest(BaseWorkerCmd):
    """
    worker rollback request
    """

    def __init__(self, actor_id, exception_msg):
        super().__init__(actor_id)
        self.__exception_msg = exception_msg

    def exception_msg(self):
        return self.__exception_msg



import gym
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, \
    TYPE_CHECKING, Union

from ray.rllib.models.catalog import ModelCatalog
from ray.rllib.models.jax.jax_modelv2 import JAXModelV2
from ray.rllib.models.modelv2 import ModelV2
from ray.rllib.models.torch.torch_action_dist import TorchDistributionWrapper
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.policy.policy import Policy, LEARNER_STATS_KEY
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.policy.torch_policy import TorchPolicy
from ray.rllib.utils import add_mixins, force_list, NullContextManager
from ray.rllib.utils.annotations import override, DeveloperAPI
from ray.rllib.utils.framework import try_import_torch, try_import_jax
from ray.rllib.utils.torch_ops import convert_to_non_torch_type
from ray.rllib.utils.typing import ModelGradients, TensorType, \
    TrainerConfigDict

if TYPE_CHECKING:
    from ray.rllib.evaluation import MultiAgentEpisode  # noqa

jax, _ = try_import_jax()
torch, _ = try_import_torch()


# TODO: (sven) Unify this with `build_tf_policy` as well.
@DeveloperAPI
def build_policy_class(
        name: str,
        framework: str,
        *,
        loss_fn: Optional[Callable[[
            Policy, ModelV2, Type[TorchDistributionWrapper], SampleBatch
        ], Union[TensorType, List[TensorType]]]],
        get_default_config: Optional[Callable[[], TrainerConfigDict]] = None,
        stats_fn: Optional[Callable[[Policy, SampleBatch], Dict[
            str, TensorType]]] = None,
        postprocess_fn: Optional[Callable[[
            Policy, SampleBatch, Optional[Dict[Any, SampleBatch]], Optional[
                "MultiAgentEpisode"]
        ], SampleBatch]] = None,
        extra_action_out_fn: Optional[Callable[[
            Policy, Dict[str, TensorType], List[TensorType], ModelV2,
            TorchDistributionWrapper
        ], Dict[str, TensorType]]] = None,
        extra_grad_process_fn: Optional[Callable[[
            Policy, "torch.optim.Optimizer", TensorType
        ], Dict[str, TensorType]]] = None,
        # TODO: (sven) Replace "fetches" with "process".
        extra_learn_fetches_fn: Optional[Callable[[Policy], Dict[
            str, TensorType]]] = None,
        optimizer_fn: Optional[Callable[[Policy, TrainerConfigDict],
                                        "torch.optim.Optimizer"]] = None,
        validate_spaces: Optional[Callable[
            [Policy, gym.Space, gym.Space, TrainerConfigDict], None]] = None,
        before_init: Optional[Callable[
            [Policy, gym.Space, gym.Space, TrainerConfigDict], None]] = None,
        before_loss_init: Optional[Callable[[
            Policy, gym.spaces.Space, gym.spaces.Space, TrainerConfigDict
        ], None]] = None,
        after_init: Optional[Callable[
            [Policy, gym.Space, gym.Space, TrainerConfigDict], None]] = None,
        _after_loss_init: Optional[Callable[[
            Policy, gym.spaces.Space, gym.spaces.Space, TrainerConfigDict
        ], None]] = None,
        action_sampler_fn: Optional[Callable[[TensorType, List[
            TensorType]], Tuple[TensorType, TensorType]]] = None,
        action_distribution_fn: Optional[Callable[[
            Policy, ModelV2, TensorType, TensorType, TensorType
        ], Tuple[TensorType, type, List[TensorType]]]] = None,
        make_model: Optional[Callable[[
            Policy, gym.spaces.Space, gym.spaces.Space, TrainerConfigDict
        ], ModelV2]] = None,
        make_model_and_action_dist: Optional[Callable[[
            Policy, gym.spaces.Space, gym.spaces.Space, TrainerConfigDict
        ], Tuple[ModelV2, Type[TorchDistributionWrapper]]]] = None,
        compute_gradients_fn: Optional[Callable[[Policy, SampleBatch], Tuple[
            ModelGradients, dict]]] = None,
        apply_gradients_fn: Optional[Callable[
            [Policy, "torch.optim.Optimizer"], None]] = None,
        mixins: Optional[List[type]] = None,
        get_batch_divisibility_req: Optional[Callable[[Policy], int]] = None
) -> Type[TorchPolicy]:
    """Helper function for creating a new Policy class at runtime.

    Supports frameworks JAX and PyTorch.

    Args:
        name (str): name of the policy (e.g., "PPOTorchPolicy")
        framework (str): Either "jax" or "torch".
        loss_fn (Optional[Callable[[Policy, ModelV2,
            Type[TorchDistributionWrapper], SampleBatch], Union[TensorType,
            List[TensorType]]]]): Callable that returns a loss tensor.
        get_default_config (Optional[Callable[[None], TrainerConfigDict]]):
            Optional callable that returns the default config to merge with any
            overrides. If None, uses only(!) the user-provided
            PartialTrainerConfigDict as dict for this Policy.
        postprocess_fn (Optional[Callable[[Policy, SampleBatch,
            Optional[Dict[Any, SampleBatch]], Optional["MultiAgentEpisode"]],
            SampleBatch]]): Optional callable for post-processing experience
            batches (called after the super's `postprocess_trajectory` method).
        stats_fn (Optional[Callable[[Policy, SampleBatch],
            Dict[str, TensorType]]]): Optional callable that returns a dict of
            values given the policy and training batch. If None,
            will use `TorchPolicy.extra_grad_info()` instead. The stats dict is
            used for logging (e.g. in TensorBoard).
        extra_action_out_fn (Optional[Callable[[Policy, Dict[str, TensorType],
            List[TensorType], ModelV2, TorchDistributionWrapper]], Dict[str,
            TensorType]]]): Optional callable that returns a dict of extra
            values to include in experiences. If None, no extra computations
            will be performed.
        extra_grad_process_fn (Optional[Callable[[Policy,
            "torch.optim.Optimizer", TensorType], Dict[str, TensorType]]]):
            Optional callable that is called after gradients are computed and
            returns a processing info dict. If None, will call the
            `TorchPolicy.extra_grad_process()` method instead.
        # TODO: (sven) dissolve naming mismatch between "learn" and "compute.."
        extra_learn_fetches_fn (Optional[Callable[[Policy],
            Dict[str, TensorType]]]): Optional callable that returns a dict of
            extra tensors from the policy after loss evaluation. If None,
            will call the `TorchPolicy.extra_compute_grad_fetches()` method
            instead.
        optimizer_fn (Optional[Callable[[Policy, TrainerConfigDict],
            "torch.optim.Optimizer"]]): Optional callable that returns a
            torch optimizer given the policy and config. If None, will call
            the `TorchPolicy.optimizer()` method instead (which returns a
            torch Adam optimizer).
        validate_spaces (Optional[Callable[[Policy, gym.Space, gym.Space,
            TrainerConfigDict], None]]): Optional callable that takes the
            Policy, observation_space, action_space, and config to check for
            correctness. If None, no spaces checking will be done.
        before_init (Optional[Callable[[Policy, gym.Space, gym.Space,
            TrainerConfigDict], None]]): Optional callable to run at the
            beginning of `Policy.__init__` that takes the same arguments as
            the Policy constructor. If None, this step will be skipped.
        before_loss_init (Optional[Callable[[Policy, gym.spaces.Space,
            gym.spaces.Space, TrainerConfigDict], None]]): Optional callable to
            run prior to loss init. If None, this step will be skipped.
        after_init (Optional[Callable[[Policy, gym.Space, gym.Space,
            TrainerConfigDict], None]]): DEPRECATED: Use `before_loss_init`
            instead.
        _after_loss_init (Optional[Callable[[Policy, gym.spaces.Space,
            gym.spaces.Space, TrainerConfigDict], None]]): Optional callable to
            run after the loss init. If None, this step will be skipped.
            This will be deprecated at some point and renamed into `after_init`
            to match `build_tf_policy()` behavior.
        action_sampler_fn (Optional[Callable[[TensorType, List[TensorType]],
            Tuple[TensorType, TensorType]]]): Optional callable returning a
            sampled action and its log-likelihood given some (obs and state)
            inputs. If None, will either use `action_distribution_fn` or
            compute actions by calling self.model, then sampling from the
            so parameterized action distribution.
        action_distribution_fn (Optional[Callable[[Policy, ModelV2, TensorType,
            TensorType, TensorType], Tuple[TensorType,
            Type[TorchDistributionWrapper], List[TensorType]]]]): A callable
            that takes the Policy, Model, the observation batch, an
            explore-flag, a timestep, and an is_training flag and returns a
            tuple of a) distribution inputs (parameters), b) a dist-class to
            generate an action distribution object from, and c) internal-state
            outputs (empty list if not applicable). If None, will either use
            `action_sampler_fn` or compute actions by calling self.model,
            then sampling from the parameterized action distribution.
        make_model (Optional[Callable[[Policy, gym.spaces.Space,
            gym.spaces.Space, TrainerConfigDict], ModelV2]]): Optional callable
            that takes the same arguments as Policy.__init__ and returns a
            model instance. The distribution class will be determined
            automatically. Note: Only one of `make_model` or
            `make_model_and_action_dist` should be provided. If both are None,
            a default Model will be created.
        make_model_and_action_dist (Optional[Callable[[Policy,
            gym.spaces.Space, gym.spaces.Space, TrainerConfigDict],
            Tuple[ModelV2, Type[TorchDistributionWrapper]]]]): Optional
            callable that takes the same arguments as Policy.__init__ and
            returns a tuple of model instance and torch action distribution
            class.
            Note: Only one of `make_model` or `make_model_and_action_dist`
            should be provided. If both are None, a default Model will be
            created.
        compute_gradients_fn (Optional[Callable[
            [Policy, SampleBatch], Tuple[ModelGradients, dict]]]): Optional
            callable that the sampled batch an computes the gradients w.r.
            to the loss function.
            If None, will call the `TorchPolicy.compute_gradients()` method
            instead.
        apply_gradients_fn (Optional[Callable[[Policy,
            "torch.optim.Optimizer"], None]]): Optional callable that
            takes a grads list and applies these to the Model's parameters.
            If None, will call the `TorchPolicy.apply_gradients()` method
            instead.
        mixins (Optional[List[type]]): Optional list of any class mixins for
            the returned policy class. These mixins will be applied in order
            and will have higher precedence than the TorchPolicy class.
        get_batch_divisibility_req (Optional[Callable[[Policy], int]]):
            Optional callable that returns the divisibility requirement for
            sample batches. If None, will assume a value of 1.

    Returns:
        Type[TorchPolicy]: TorchPolicy child class constructed from the
            specified args.
    """

    original_kwargs = locals().copy()
    parent_cls = TorchPolicy
    base = add_mixins(parent_cls, mixins)

    class policy_cls(base):
        def __init__(self, obs_space, action_space, config):
            # Set up the config from possible default-config fn and given
            # config arg.
            if get_default_config:
                config = dict(get_default_config(), **config)
            self.config = config

            # Set the DL framework for this Policy.
            self.framework = self.config["framework"] = framework

            # Validate observation- and action-spaces.
            if validate_spaces:
                validate_spaces(self, obs_space, action_space, self.config)

            # Do some pre-initialization steps.
            if before_init:
                before_init(self, obs_space, action_space, self.config)

            # Model is customized (use default action dist class).
            if make_model:
                assert make_model_and_action_dist is None, \
                    "Either `make_model` or `make_model_and_action_dist`" \
                    " must be None!"
                self.model = make_model(self, obs_space, action_space, config)
                dist_class, _ = ModelCatalog.get_action_dist(
                    action_space, self.config["model"], framework=framework)
            # Model and action dist class are customized.
            elif make_model_and_action_dist:
                self.model, dist_class = make_model_and_action_dist(
                    self, obs_space, action_space, config)
            # Use default model and default action dist.
            else:
                dist_class, logit_dim = ModelCatalog.get_action_dist(
                    action_space, self.config["model"], framework=framework)
                self.model = ModelCatalog.get_model_v2(
                    obs_space=obs_space,
                    action_space=action_space,
                    num_outputs=logit_dim,
                    model_config=self.config["model"],
                    framework=framework)

            # Make sure, we passed in a correct Model factory.
            model_cls = TorchModelV2 if framework == "torch" else JAXModelV2
            assert isinstance(self.model, model_cls), \
                "ERROR: Generated Model must be a TorchModelV2 object!"

            # Call the framework-specific Policy constructor.
            self.parent_cls = parent_cls
            self.parent_cls.__init__(
                self,
                observation_space=obs_space,
                action_space=action_space,
                config=config,
                model=self.model,
                loss=loss_fn,
                action_distribution_class=dist_class,
                action_sampler_fn=action_sampler_fn,
                action_distribution_fn=action_distribution_fn,
                max_seq_len=config["model"]["max_seq_len"],
                get_batch_divisibility_req=get_batch_divisibility_req,
            )

            # Merge Model's view requirements into Policy's.
            self.view_requirements.update(self.model.view_requirements)

            _before_loss_init = before_loss_init or after_init
            if _before_loss_init:
                _before_loss_init(self, self.observation_space,
                                  self.action_space, config)

            # Perform test runs through postprocessing- and loss functions.
            self._initialize_loss_from_dummy_batch(
                auto_remove_unneeded_view_reqs=True,
                stats_fn=stats_fn,
            )

            if _after_loss_init:
                _after_loss_init(self, obs_space, action_space, config)

            # Got to reset global_timestep again after this fake run-through.
            self.global_timestep = 0

        @override(Policy)
        def postprocess_trajectory(self,
                                   sample_batch,
                                   other_agent_batches=None,
                                   episode=None):
            # Do all post-processing always with no_grad().
            # Not using this here will introduce a memory leak
            # in torch (issue #6962).
            with self._no_grad_context():
                # Call super's postprocess_trajectory first.
                sample_batch = super().postprocess_trajectory(
                    sample_batch, other_agent_batches, episode)
                if postprocess_fn:
                    return postprocess_fn(self, sample_batch,
                                          other_agent_batches, episode)

                return sample_batch

        @override(parent_cls)
        def extra_grad_process(self, optimizer, loss):
            """Called after optimizer.zero_grad() and loss.backward() calls.

            Allows for gradient processing before optimizer.step() is called.
            E.g. for gradient clipping.
            """
            if extra_grad_process_fn:
                return extra_grad_process_fn(self, optimizer, loss)
            else:
                return parent_cls.extra_grad_process(self, optimizer, loss)

        @override(parent_cls)
        def extra_compute_grad_fetches(self):
            if extra_learn_fetches_fn:
                fetches = convert_to_non_torch_type(
                    extra_learn_fetches_fn(self))
                # Auto-add empty learner stats dict if needed.
                return dict({LEARNER_STATS_KEY: {}}, **fetches)
            else:
                return parent_cls.extra_compute_grad_fetches(self)

        @override(parent_cls)
        def compute_gradients(self, batch):
            if compute_gradients_fn:
                return compute_gradients_fn(self, batch)
            else:
                return parent_cls.compute_gradients(self, batch)

        @override(parent_cls)
        def apply_gradients(self, gradients):
            if apply_gradients_fn:
                apply_gradients_fn(self, gradients)
            else:
                parent_cls.apply_gradients(self, gradients)

        @override(parent_cls)
        def extra_action_out(self, input_dict, state_batches, model,
                             action_dist):
            with self._no_grad_context():
                if extra_action_out_fn:
                    stats_dict = extra_action_out_fn(
                        self, input_dict, state_batches, model, action_dist)
                else:
                    stats_dict = parent_cls.extra_action_out(
                        self, input_dict, state_batches, model, action_dist)
                return self._convert_to_non_torch_type(stats_dict)

        @override(parent_cls)
        def optimizer(self):
            if optimizer_fn:
                optimizers = optimizer_fn(self, self.config)
            else:
                optimizers = parent_cls.optimizer(self)
            optimizers = force_list(optimizers)
            if getattr(self, "exploration", None):
                optimizers = self.exploration.get_exploration_optimizer(
                    optimizers)
            return optimizers

        @override(parent_cls)
        def extra_grad_info(self, train_batch):
            with self._no_grad_context():
                if stats_fn:
                    stats_dict = stats_fn(self, train_batch)
                else:
                    stats_dict = self.parent_cls.extra_grad_info(
                        self, train_batch)
                return self._convert_to_non_torch_type(stats_dict)

        def _no_grad_context(self):
            if self.framework == "torch":
                return torch.no_grad()
            return NullContextManager()

        def _convert_to_non_torch_type(self, data):
            if self.framework == "torch":
                return convert_to_non_torch_type(data)
            return data

    def with_updates(**overrides):
        """Creates a Torch|JAXPolicy cls based on settings of another one.

        Keyword Args:
            **overrides: The settings (passed into `build_torch_policy`) that
                should be different from the class that this method is called
                on.

        Returns:
            type: A new Torch|JAXPolicy sub-class.

        Examples:
        >> MySpecialDQNPolicyClass = DQNTorchPolicy.with_updates(
        ..    name="MySpecialDQNPolicyClass",
        ..    loss_function=[some_new_loss_function],
        .. )
        """
        return build_policy_class(**dict(original_kwargs, **overrides))

    policy_cls.with_updates = staticmethod(with_updates)
    policy_cls.__name__ = name
    policy_cls.__qualname__ = name
    return policy_cls



import math
import time
from abc import ABC
from collections import defaultdict
from collections.abc import Iterable
from enum import Enum
import os
from typing import Any, Dict, List, Optional, Tuple

import ray
from ray import cloudpickle
from ray.actor import ActorHandle
from ray.serve.async_goal_manager import AsyncGoalManager
from ray.serve.common import (BackendInfo, BackendTag, Duration, GoalId,
                              ReplicaTag)
from ray.serve.config import BackendConfig
from ray.serve.constants import RESERVED_VERSION_TAG
from ray.serve.kv_store import RayInternalKVStore
from ray.serve.long_poll import LongPollHost, LongPollNamespace
from ray.serve.utils import format_actor_name, get_random_letters, logger

CHECKPOINT_KEY = "serve-backend-state-checkpoint"
SLOW_STARTUP_WARNING_S = 30
SLOW_STARTUP_WARNING_PERIOD_S = 30


class ReplicaState(Enum):
    SHOULD_START = 1
    STARTING_OR_UPDATING = 2
    RUNNING = 3
    SHOULD_STOP = 4
    STOPPING = 5


ALL_REPLICA_STATES = list(ReplicaState)
USE_PLACEMENT_GROUP = os.environ.get("SERVE_USE_PLACEMENT_GROUP", "1") != "0"


class ActorReplicaWrapper:
    """Wraps a Ray actor for a backend replica.

    This is primarily defined so that we can mock out actual Ray operations
    for unit testing.

    *All Ray API calls should be made here, not in BackendState.*
    """

    def __init__(self, actor_name: str, detached: bool, controller_name: str,
                 replica_tag: ReplicaTag, backend_tag: BackendTag):
        self._actor_name = actor_name
        self._placement_group_name = self._actor_name + "_placement_group"
        self._detached = detached
        self._controller_name = controller_name
        self._replica_tag = replica_tag
        self._backend_tag = backend_tag

        self._ready_obj_ref = None
        self._drain_obj_ref = None
        self._actor_resources = None
        self._health_check_ref = None

        # Storing the handles is necessary to keep the actor and PG alive in
        # the non-detached case.
        self._actor_handle = None
        self._placement_group = None

    def __get_state__(self) -> Dict[Any, Any]:
        clean_dict = self.__dict__.copy()
        del clean_dict["_ready_obj_ref"]
        del clean_dict["_drain_obj_ref"]
        return clean_dict

    def __set_state__(self, d: Dict[Any, Any]) -> None:
        self.__dict__ = d
        self._ready_obj_ref = None
        self._drain_obj_ref = None

    @property
    def actor_handle(self) -> ActorHandle:
        return ray.get_actor(self._actor_name)

    def start_or_update(self, backend_info: BackendInfo):
        self._actor_resources = backend_info.replica_config.resource_dict

        # Feature flagging because of placement groups doesn't handle
        # newly added nodes.
        # https://github.com/ray-project/ray/issues/15801
        if USE_PLACEMENT_GROUP:
            try:
                self._placement_group = ray.util.get_placement_group(
                    self._placement_group_name)
            except ValueError:
                logger.debug(
                    "Creating placement group '{}' for backend '{}'".format(
                        self._placement_group_name, self._backend_tag))
                self._placement_group = ray.util.placement_group(
                    [self._actor_resources],
                    lifetime="detached" if self._detached else None,
                    name=self._placement_group_name)

        try:
            self._actor_handle = ray.get_actor(self._actor_name)
        except ValueError:
            logger.debug("Starting replica '{}' for backend '{}'.".format(
                self._replica_tag, self._backend_tag))
            self._actor_handle = backend_info.actor_def.options(
                name=self._actor_name,
                lifetime="detached" if self._detached else None,
                placement_group=self._placement_group,
                placement_group_capture_child_tasks=False,
                **backend_info.replica_config.ray_actor_options).remote(
                    self._backend_tag, self._replica_tag,
                    backend_info.replica_config.init_args,
                    backend_info.backend_config, self._controller_name)

        self._ready_obj_ref = self._actor_handle.reconfigure.remote(
            backend_info.backend_config.user_config)

    def check_ready(self) -> bool:
        ready, _ = ray.wait([self._ready_obj_ref], timeout=0)
        return len(ready) == 1

    @property
    def actor_resources(self) -> Dict[str, float]:
        return self._actor_resources

    def graceful_stop(self) -> None:
        """Request the actor to exit gracefully."""
        try:
            handle = ray.get_actor(self._actor_name)
            self._drain_obj_ref = handle.drain_pending_queries.remote()
        except ValueError:
            pass

    def check_stopped(self) -> bool:
        """Check if the actor has exited."""
        try:
            handle = ray.get_actor(self._actor_name)
            ready, _ = ray.wait([self._drain_obj_ref], timeout=0)
            stopped = len(ready) == 1
            if stopped:
                ray.kill(handle, no_restart=True)
        except ValueError:
            stopped = True

        return stopped

    def check_health(self) -> bool:
        """Check if the actor is healthy."""
        if self._health_check_ref is None:
            self._health_check_ref = self._actor_handle.run_forever.remote()

        ready, _ = ray.wait([self._health_check_ref], timeout=0)
        return len(ready) == 0

    def force_stop(self):
        """Force the actor to exit without shutting down gracefully."""
        try:
            ray.kill(ray.get_actor(self._actor_name))
        except ValueError:
            pass

    def cleanup(self):
        """Clean up any remaining resources after the actor has exited.

        Currently, this just removes the placement group.
        """
        if not USE_PLACEMENT_GROUP:
            return

        try:
            ray.util.remove_placement_group(
                ray.util.get_placement_group(self._placement_group_name))
        except ValueError:
            pass


class BackendVersion:
    def __init__(self,
                 code_version: Optional[str],
                 user_config: Optional[Any] = None):
        if code_version is not None and not isinstance(code_version, str):
            raise TypeError(
                f"code_version must be str, got {type(code_version)}.")

        self.code_version = code_version
        self.user_config_hash = self._hash_user_config(user_config)
        self._hash = hash((self.code_version, self.user_config_hash))

    def _hash_user_config(self, user_config: Any) -> int:
        """Hash the user config.

        We want users to be able to pass lists and dictionaries for
        convenience, but these are not hashable types because they're mutable.

        This supports lists and dictionaries by recursively converting them
        into immutable tuples and then hashing them.
        """
        try:
            return hash(user_config)
        except TypeError:
            pass

        if isinstance(user_config, dict):
            keys = tuple(sorted(user_config))
            val_hashes = tuple(
                self._hash_user_config(user_config[k]) for k in keys)
            return hash((hash(keys), hash(val_hashes)))
        elif isinstance(user_config, Iterable):
            return hash(
                tuple(self._hash_user_config(item) for item in user_config))
        else:
            raise TypeError(
                "user_config must contain only lists, dicts, or hashable "
                f"types. Got {type(user_config)}.")

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: Any) -> bool:
        return self._hash == other._hash


class VersionedReplica(ABC):
    @property
    def version(self) -> BackendVersion:
        pass


class BackendReplica(VersionedReplica):
    """Manages state transitions for backend replicas.

    This is basically a checkpointable lightweight state machine.
    """

    def __init__(self, controller_name: str, detached: bool,
                 replica_tag: ReplicaTag, backend_tag: BackendTag,
                 version: BackendVersion):
        self._actor = ActorReplicaWrapper(
            format_actor_name(replica_tag, controller_name), detached,
            controller_name, replica_tag, backend_tag)
        self._controller_name = controller_name
        self._replica_tag = replica_tag
        self._backend_tag = backend_tag
        self._version = version
        self._start_time = None
        self._prev_slow_startup_warning_time = None
        self._state = ReplicaState.SHOULD_START

    def __get_state__(self) -> Dict[Any, Any]:
        return self.__dict__.copy()

    def __set_state__(self, d: Dict[Any, Any]) -> None:
        self.__dict__ = d
        self._recover_from_checkpoint()

    def _recover_from_checkpoint(self) -> None:
        if self._state == ReplicaState.STARTING_OR_UPDATING:
            # We do not need to pass in the class here because the actor
            # creation has already been started if this class was checkpointed
            # in the STARTING_OR_UPDATING state.
            self.start_or_update()
        elif self._state == ReplicaState.STOPPING:
            self.stop()

    @property
    def replica_tag(self) -> ReplicaTag:
        return self._replica_tag

    @property
    def version(self):
        return self._version

    @property
    def actor_handle(self) -> ActorHandle:
        assert self._state is not ReplicaState.SHOULD_START, (
            f"State must not be {ReplicaState.SHOULD_START}")
        return self._actor.actor_handle

    def start_or_update(self, backend_info: BackendInfo,
                        version: BackendVersion) -> None:
        """Transition from SHOULD_START -> STARTING_OR_UPDATING.

        Should handle the case where it's already STARTING_OR_UPDATING.
        """
        assert self._state in {
            ReplicaState.SHOULD_START,
            ReplicaState.STARTING_OR_UPDATING,
            ReplicaState.RUNNING,
        }, (f"State must be {ReplicaState.SHOULD_START}, "
            f"{ReplicaState.STARTING_OR_UPDATING}, or "
            f"{ReplicaState.RUNNING} *not* {self._state}.")

        self._actor.start_or_update(backend_info)
        self._start_time = time.time()
        self._prev_slow_startup_warning_time = time.time()
        self._state = ReplicaState.STARTING_OR_UPDATING
        self._version = version

    def check_started(self) -> Tuple[bool, bool]:
        """Check if the replica has started. If so, transition to RUNNING.

        Should handle the case where the replica has already stopped.

        Returns:
            (ready, past_slow_startup_threshold)
        """
        if self._state == ReplicaState.RUNNING:
            return True, False
        assert self._state == ReplicaState.STARTING_OR_UPDATING, (
            f"State must be {ReplicaState.STARTING_OR_UPDATING}, "
            f"*not* {self._state}.")

        if self._actor.check_ready():
            self._state = ReplicaState.RUNNING
            return True, False

        time_since_start = time.time() - self._start_time
        return False, time_since_start > SLOW_STARTUP_WARNING_S

    def set_should_stop(self, graceful_shutdown_timeout_s: Duration) -> None:
        """Mark the replica to be stopped in the future.

        Should handle the case where the replica has already been marked to
        stop.
        """
        self._state = ReplicaState.SHOULD_STOP
        self._graceful_shutdown_timeout_s = graceful_shutdown_timeout_s

    def stop(self) -> None:
        """Stop the replica.

        Should handle the case where the replica is already stopped.
        """
        # We need to handle transitions from:
        #  SHOULD_START -> SHOULD_STOP -> STOPPING
        # This means that the replica_handle may not have been created.

        assert self._state in {
            ReplicaState.SHOULD_STOP, ReplicaState.STOPPING
        }, (f"State must be {ReplicaState.SHOULD_STOP} or "
            f"{ReplicaState.STOPPING}, *not* {self._state}")

        self._actor.graceful_stop()
        self._state = ReplicaState.STOPPING
        self._shutdown_deadline = time.time(
        ) + self._graceful_shutdown_timeout_s

    def check_stopped(self) -> bool:
        """Check if the replica has finished stopping."""
        assert self._state == ReplicaState.STOPPING, (
            f"State must be {ReplicaState.STOPPING}, *not* {self._state}")

        if self._actor.check_stopped():
            # Clean up any associated resources (e.g., placement group).
            self._actor.cleanup()
            return True

        timeout_passed = time.time() >= self._shutdown_deadline

        if timeout_passed:
            # Graceful period passed, kill it forcefully.
            # This will be called repeatedly until the replica shuts down.
            logger.debug(
                f"Replica {self._replica_tag} did not shutdown after "
                f"{self._graceful_shutdown_timeout_s}s, force-killing.")

            self._actor.force_stop()
        return False

    def check_health(self) -> bool:
        """Check if the replica is still alive.

        Returns `True` if the replica is healthy, else `False`.
        """
        return self._actor.check_health()

    def resource_requirements(
            self) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Returns required and currently available resources.

        Only resources with nonzero requirements will be included in the
        required dict and only resources in the required dict will be
        included in the available dict (filtered for relevance).
        """
        required = {
            k: v
            for k, v in self._actor.actor_resources.items() if v > 0
        }
        available = {
            k: v
            for k, v in ray.available_resources().items() if k in required
        }
        return required, available


class ReplicaStateContainer:
    """Container for mapping ReplicaStates to lists of BackendReplicas."""

    def __init__(self):
        self._replicas: Dict[ReplicaState, List[BackendReplica]] = defaultdict(
            list)

    def add(self, state: ReplicaState, replica: VersionedReplica):
        """Add the provided replica under the provided state.

        Args:
            state (ReplicaState): state to add the replica under.
            replica (VersionedReplica): replica to add.
        """
        assert isinstance(state, ReplicaState)
        assert isinstance(replica, VersionedReplica)
        self._replicas[state].append(replica)

    def get(self, states: Optional[List[ReplicaState]] = None
            ) -> List[BackendReplica]:
        """Get all replicas of the given states.

        This does not remove them from the container. Replicas are returned
        in order of state as passed in.

        Args:
            states (str): states to consider. If not specified, all replicas
                are considered.
        """
        if states is None:
            states = ALL_REPLICA_STATES

        assert isinstance(states, list)

        return sum((self._replicas[state] for state in states), [])

    def pop(self,
            exclude_version: Optional[BackendVersion] = None,
            states: Optional[List[ReplicaState]] = None,
            max_replicas: Optional[int] = math.inf) -> List[VersionedReplica]:
        """Get and remove all replicas of the given states.

        This removes the replicas from the container. Replicas are returned
        in order of state as passed in.

        Args:
            exclude_version (BackendVersion): if specified, replicas of the
                provided version will *not* be removed.
            states (str): states to consider. If not specified, all replicas
                are considered.
            max_replicas (int): max number of replicas to return. If not
                specified, will pop all replicas matching the criteria.
        """
        if states is None:
            states = ALL_REPLICA_STATES

        assert (exclude_version is None
                or isinstance(exclude_version, BackendVersion))
        assert isinstance(states, list)

        replicas = []
        for state in states:
            popped = []
            remaining = []
            for replica in self._replicas[state]:
                if len(replicas) + len(popped) == max_replicas:
                    remaining.append(replica)
                elif (exclude_version is not None
                      and replica.version == exclude_version):
                    remaining.append(replica)
                else:
                    popped.append(replica)
            self._replicas[state] = remaining
            replicas.extend(popped)

        return replicas

    def count(self,
              exclude_version: Optional[BackendVersion] = None,
              version: Optional[BackendVersion] = None,
              states: Optional[List[ReplicaState]] = None):
        """Get the total count of replicas of the given states.

        Args:
            exclude_version(BackendVersion): version to exclude. If not
                specified, all versions are considered.
            version(BackendVersion): version to filter to. If not specified,
                all versions are considered.
            states (str): states to consider. If not specified, all replicas
                are considered.
        """
        if states is None:
            states = ALL_REPLICA_STATES
        assert isinstance(states, list)
        assert (exclude_version is None
                or isinstance(exclude_version, BackendVersion))
        assert version is None or isinstance(version, BackendVersion)
        if exclude_version is None and version is None:
            return sum(len(self._replicas[state]) for state in states)
        elif exclude_version is None and version is not None:
            return sum(
                len(
                    list(
                        filter(lambda r: r.version == version, self._replicas[
                            state]))) for state in states)
        elif exclude_version is not None and version is None:
            return sum(
                len(
                    list(
                        filter(lambda r: r.version != exclude_version,
                               self._replicas[state]))) for state in states)
        else:
            raise ValueError(
                "Only one of `version` or `exclude_version` may be provided.")

    def __str__(self):
        return str(self._replicas)

    def __repr__(self):
        return repr(self._replicas)


class BackendState:
    """Manages all state for backends in the system.

    This class is *not* thread safe, so any state-modifying methods should be
    called with a lock held.
    """

    def __init__(self, controller_name: str, detached: bool,
                 kv_store: RayInternalKVStore, long_poll_host: LongPollHost,
                 goal_manager: AsyncGoalManager):

        self._controller_name = controller_name
        self._detached = detached
        self._kv_store = kv_store
        self._long_poll_host = long_poll_host
        self._goal_manager = goal_manager
        self._replicas: Dict[BackendTag, ReplicaStateContainer] = dict()
        self._backend_metadata: Dict[BackendTag, BackendInfo] = dict()
        self._target_replicas: Dict[BackendTag, int] = defaultdict(int)
        self._backend_goals: Dict[BackendTag, GoalId] = dict()
        self._target_versions: Dict[BackendTag, BackendVersion] = dict()
        self._prev_startup_warnings: Dict[BackendTag, float] = defaultdict(
            float)

        checkpoint = self._kv_store.get(CHECKPOINT_KEY)
        if checkpoint is not None:
            (self._replicas, self._backend_metadata, self._target_replicas,
             self._target_versions,
             self._backend_goals) = cloudpickle.loads(checkpoint)

            for goal_id in self._backend_goals.values():
                self._goal_manager.create_goal(goal_id)

        self._notify_backend_configs_changed()
        self._notify_replica_handles_changed()

    def shutdown(self) -> List[GoalId]:
        """
        Shutdown all running replicas by notifying the controller, and leave
        it to the controller event loop to take actions afterwards.

        Once shutdown signal is received, it will also prevent any new
        deployments or replicas from being created.

        One can send multiple shutdown signals but won't effectively make any
        difference compare to calling it once.
        """

        shutdown_goals = []
        for backend_tag, _ in self._replicas.items():
            goal = self.delete_backend(backend_tag, force_kill=True)
            if goal is not None:
                shutdown_goals.append(goal)

        # TODO(jiaodong): This might not be 100% safe since we deleted
        # everything without ensuring all shutdown goals are completed
        # yet. Need to address in follow-up PRs.
        self._kv_store.delete(CHECKPOINT_KEY)

        # TODO(jiaodong): Need to add some logic to prevent new replicas
        # from being created once shutdown signal is sent.
        return shutdown_goals

    def _checkpoint(self) -> None:
        self._kv_store.put(
            CHECKPOINT_KEY,
            cloudpickle.dumps(
                (self._replicas, self._backend_metadata, self._target_replicas,
                 self._target_versions, self._backend_goals)))

    def _notify_backend_configs_changed(
            self, key: Optional[BackendTag] = None) -> None:
        for key, config in self.get_backend_configs(key).items():
            self._long_poll_host.notify_changed(
                (LongPollNamespace.BACKEND_CONFIGS, key),
                config,
            )

    def get_running_replica_handles(
            self,
            filter_tag: Optional[BackendTag] = None,
    ) -> Dict[BackendTag, Dict[ReplicaTag, ActorHandle]]:
        return {
            backend_tag: {
                backend_replica.replica_tag: backend_replica.actor_handle
                for backend_replica in replicas_container.get(
                    [ReplicaState.RUNNING])
            }
            for backend_tag, replicas_container in self._replicas.items()
            if filter_tag is None or backend_tag == filter_tag
        }

    def _notify_replica_handles_changed(
            self, key: Optional[BackendTag] = None) -> None:
        for key, replica_dict in self.get_running_replica_handles(key).items():
            self._long_poll_host.notify_changed(
                (LongPollNamespace.REPLICA_HANDLES, key),
                list(replica_dict.values()),
            )

    def get_backend_configs(
            self,
            filter_tag: Optional[BackendTag] = None,
    ) -> Dict[BackendTag, BackendConfig]:
        return {
            tag: info.backend_config
            for tag, info in self._backend_metadata.items()
            if filter_tag is None or tag == filter_tag
        }

    def get_backend(self, backend_tag: BackendTag) -> Optional[BackendInfo]:
        return self._backend_metadata.get(backend_tag)

    def _set_backend_goal(self, backend_tag: BackendTag,
                          backend_info: Optional[BackendInfo]) -> None:
        """
        Set desirable state for a given backend, identified by tag.

        Args:
            backend_tag (BackendTag): Identifier of a backend
            backend_info (Optional[BackendInfo]): Contains backend and
                replica config, if passed in as None, we're marking
                target backend as shutting down.
        """
        existing_goal_id = self._backend_goals.get(backend_tag)
        new_goal_id = self._goal_manager.create_goal()

        if backend_info is not None:
            self._backend_metadata[backend_tag] = backend_info
            self._target_replicas[
                backend_tag] = backend_info.backend_config.num_replicas

            if backend_info.version is not None:
                code_version = backend_info.version
            else:
                code_version = get_random_letters()

            self._target_versions[backend_tag] = BackendVersion(
                code_version,
                user_config=backend_info.backend_config.user_config)

        else:
            self._target_replicas[backend_tag] = 0

        self._backend_goals[backend_tag] = new_goal_id

        return new_goal_id, existing_goal_id

    def deploy_backend(self, backend_tag: BackendTag, backend_info: BackendInfo
                       ) -> Tuple[Optional[GoalId], bool]:
        """Deploy the backend.

        If the backend already exists with the same version, this is a no-op
        and returns the GoalId corresponding to the existing update if there
        is one.

        Returns:
            GoalId, bool: The GoalId for the client to wait for and whether or
            not the backend is being updated.
        """
        # Ensures this method is idempotent.
        existing_info = self._backend_metadata.get(backend_tag)
        if existing_info is not None:
            # Old codepath. We use RESERVED_VERSION_TAG to distinguish that
            # we shouldn't use versions at all to determine redeployment
            # because `None` is used to indicate always redeploying.
            if backend_info.version == RESERVED_VERSION_TAG:
                if (existing_info.backend_config == backend_info.backend_config
                        and existing_info.replica_config ==
                        backend_info.replica_config):
                    return self._backend_goals.get(backend_tag, None), False
            # New codepath: treat version as ground truth for implementation.
            elif (existing_info.backend_config == backend_info.backend_config
                  and backend_info.version is not None
                  and existing_info.version == backend_info.version):
                return self._backend_goals.get(backend_tag, None), False

        if backend_tag not in self._replicas:
            self._replicas[backend_tag] = ReplicaStateContainer()

        new_goal_id, existing_goal_id = self._set_backend_goal(
            backend_tag, backend_info)

        # NOTE(edoakes): we must write a checkpoint before starting new
        # or pushing the updated config to avoid inconsistent state if we
        # crash while making the change.
        self._checkpoint()
        self._notify_backend_configs_changed(backend_tag)

        if existing_goal_id is not None:
            self._goal_manager.complete_goal(existing_goal_id)
        return new_goal_id, True

    def delete_backend(self, backend_tag: BackendTag,
                       force_kill: bool = False) -> Optional[GoalId]:
        # This method must be idempotent. We should validate that the
        # specified backend exists on the client.
        if backend_tag not in self._backend_metadata:
            return None

        new_goal_id, existing_goal_id = self._set_backend_goal(
            backend_tag, None)
        if force_kill:
            self._backend_metadata[
                backend_tag].backend_config.\
                    experimental_graceful_shutdown_timeout_s = 0

        self._checkpoint()
        if existing_goal_id is not None:
            self._goal_manager.complete_goal(existing_goal_id)
        return new_goal_id

    def _stop_wrong_version_replicas(
            self, backend_tag: BackendTag, replicas: ReplicaStateContainer,
            target_replicas: int, target_version: str,
            graceful_shutdown_timeout_s: float) -> int:
        """Stops replicas with outdated versions to implement rolling updates.

        This includes both explicit code version updates and changes to the
        user_config.
        """
        # NOTE(edoakes): this short-circuits when using the legacy
        # `create_backend` codepath -- it can be removed once we deprecate
        # that as the version should never be None.
        if target_version is None:
            return 0

        # Short circuit if target replicas is 0 (the backend is being deleted)
        # because this will be handled in the main loop.
        if target_replicas == 0:
            return 0

        # We include SHOULD_START and STARTING_OR_UPDATING replicas here
        # because if there are replicas still pending startup, we may as well
        # terminate them and start new version replicas instead.
        old_running_replicas = replicas.count(
            exclude_version=target_version,
            states=[
                ReplicaState.SHOULD_START, ReplicaState.STARTING_OR_UPDATING,
                ReplicaState.RUNNING
            ])
        old_stopping_replicas = replicas.count(
            exclude_version=target_version,
            states=[ReplicaState.SHOULD_STOP, ReplicaState.STOPPING])
        new_running_replicas = replicas.count(
            version=target_version, states=[ReplicaState.RUNNING])

        # If the backend is currently scaling down, let the scale down
        # complete before doing a rolling update.
        if target_replicas < old_running_replicas + old_stopping_replicas:
            return 0

        # The number of replicas that are currently in transition between
        # an old version and the new version. Note that we cannot directly
        # count the number of stopping replicas because once replicas finish
        # stopping, they are removed from the data structure.
        pending_replicas = (
            target_replicas - new_running_replicas - old_running_replicas)

        # Maximum number of replicas that can be updating at any given time.
        # There should never be more than rollout_size old replicas stopping
        # or rollout_size new replicas starting.
        rollout_size = max(int(0.2 * target_replicas), 1)
        max_to_stop = max(rollout_size - pending_replicas, 0)

        replicas_to_update = replicas.pop(
            exclude_version=target_version,
            states=[
                ReplicaState.SHOULD_START, ReplicaState.STARTING_OR_UPDATING,
                ReplicaState.RUNNING
            ],
            max_replicas=max_to_stop)

        code_version_changes = 0
        user_config_changes = 0
        for replica in replicas_to_update:
            # If the code version is a mismatch, we stop the replica. A new one
            # with the correct version will be started later as part of the
            # normal scale-up process.
            if replica.version.code_version != target_version.code_version:
                code_version_changes += 1
                replica.set_should_stop(graceful_shutdown_timeout_s)
                replicas.add(ReplicaState.SHOULD_STOP, replica)
            # If only the user_config is a mismatch, we update it dynamically
            # without restarting the replica.
            elif (replica.version.user_config_hash !=
                  target_version.user_config_hash):
                user_config_changes += 1
                replica.start_or_update(self._backend_metadata[backend_tag],
                                        target_version)
                replicas.add(ReplicaState.STARTING_OR_UPDATING, replica)
            else:
                assert False, "Update must be code version or user config."

        if code_version_changes > 0:
            logger.info(f"Stopping {code_version_changes} replicas of backend "
                        f"'{backend_tag}' with outdated versions.")

        if user_config_changes > 0:
            logger.info(f"Updating {user_config_changes} replicas of backend "
                        f"'{backend_tag}' with outdated user_configs.")

        return len(replicas_to_update)

    def _scale_backend_replicas(
            self,
            backend_tag: BackendTag,
            target_replicas: int,
            target_version: str,
    ) -> bool:
        """Scale the given backend to the number of replicas.

        NOTE: this does not actually start or stop the replicas, but instead
        adds them to ReplicaState.SHOULD_START or ReplicaState.SHOULD_STOP.
        The caller is responsible for then first writing a checkpoint and then
        actually starting/stopping the intended replicas. This avoids
        inconsistencies with starting/stopping a replica and then crashing
        before writing a checkpoint.
        """
        assert (backend_tag in self._backend_metadata
                ), "Backend {} is not registered.".format(backend_tag)
        assert target_replicas >= 0, ("Number of replicas must be"
                                      " greater than or equal to 0.")

        backend_info: BackendInfo = self._backend_metadata[backend_tag]
        graceful_shutdown_timeout_s = (
            backend_info.backend_config.
            experimental_graceful_shutdown_timeout_s)

        self._stop_wrong_version_replicas(
            backend_tag, self._replicas[backend_tag], target_replicas,
            target_version, graceful_shutdown_timeout_s)

        current_replicas = self._replicas[backend_tag].count(states=[
            ReplicaState.SHOULD_START, ReplicaState.STARTING_OR_UPDATING,
            ReplicaState.RUNNING
        ])

        delta_replicas = target_replicas - current_replicas
        if delta_replicas == 0:
            return False

        elif delta_replicas > 0:
            # Don't ever exceed target_replicas.
            stopping_replicas = self._replicas[backend_tag].count(states=[
                ReplicaState.SHOULD_STOP,
                ReplicaState.STOPPING,
            ])
            to_add = max(delta_replicas - stopping_replicas, 0)
            if to_add > 0:
                logger.info(f"Adding {to_add} replicas "
                            f"to backend '{backend_tag}'.")
            for _ in range(to_add):
                replica_tag = "{}#{}".format(backend_tag, get_random_letters())
                self._replicas[backend_tag].add(
                    ReplicaState.SHOULD_START,
                    BackendReplica(self._controller_name, self._detached,
                                   replica_tag, backend_tag, target_version))

        elif delta_replicas < 0:
            to_remove = -delta_replicas
            logger.info(f"Removing {to_remove} replicas "
                        f"from backend '{backend_tag}'.")
            replicas_to_stop = self._replicas[backend_tag].pop(
                states=[
                    ReplicaState.SHOULD_START,
                    ReplicaState.STARTING_OR_UPDATING, ReplicaState.RUNNING
                ],
                max_replicas=to_remove)

            for replica in replicas_to_stop:
                replica.set_should_stop(graceful_shutdown_timeout_s)
                self._replicas[backend_tag].add(ReplicaState.SHOULD_STOP,
                                                replica)

        return True

    def _scale_all_backends(self):
        checkpoint_needed = False
        for backend_tag, num_replicas in list(self._target_replicas.items()):
            checkpoint_needed |= self._scale_backend_replicas(
                backend_tag, num_replicas, self._target_versions[backend_tag])

        if checkpoint_needed:
            self._checkpoint()

    def _completed_goals(self) -> List[GoalId]:
        completed_goals = []
        deleted_backends = []
        for backend_tag in self._replicas:
            target_count = self._target_replicas.get(backend_tag, 0)

            # If we have pending ops, the current goal is *not* ready.
            if (self._replicas[backend_tag].count(states=[
                    ReplicaState.SHOULD_START,
                    ReplicaState.STARTING_OR_UPDATING,
                    ReplicaState.SHOULD_STOP,
                    ReplicaState.STOPPING,
            ]) > 0):
                continue

            running = self._replicas[backend_tag].get(
                states=[ReplicaState.RUNNING])
            running_count = len(running)

            # Check for deleting.
            if target_count == 0 and running_count == 0:
                deleted_backends.append(backend_tag)
                completed_goals.append(
                    self._backend_goals.pop(backend_tag, None))

            # Check for a non-zero number of backends.
            elif target_count == running_count:
                # Check that all running replicas are the target version.
                target_version = self._target_versions[backend_tag]
                if all(r.version == target_version for r in running):
                    completed_goals.append(
                        self._backend_goals.pop(backend_tag, None))

        for backend_tag in deleted_backends:
            del self._replicas[backend_tag]
            del self._backend_metadata[backend_tag]
            del self._target_replicas[backend_tag]
            del self._target_versions[backend_tag]

        return [goal for goal in completed_goals if goal]

    def update(self) -> bool:
        """Updates the state of all running replicas to match the goal state.
        """
        self._scale_all_backends()

        for goal_id in self._completed_goals():
            self._goal_manager.complete_goal(goal_id)

        transitioned_backend_tags = set()
        for backend_tag, replicas in self._replicas.items():
            for replica in replicas.pop(states=[ReplicaState.RUNNING]):
                if replica.check_health():
                    replicas.add(ReplicaState.RUNNING, replica)
                else:
                    logger.warning(
                        f"Replica {replica.replica_tag} of backend "
                        f"{backend_tag} failed health check, stopping it.")
                    replica.set_should_stop(0)
                    replicas.add(ReplicaState.SHOULD_STOP, replica)

            for replica in replicas.pop(states=[ReplicaState.SHOULD_START]):
                replica.start_or_update(self._backend_metadata[backend_tag],
                                        self._target_versions[backend_tag])
                replicas.add(ReplicaState.STARTING_OR_UPDATING, replica)

            for replica in replicas.pop(states=[ReplicaState.SHOULD_STOP]):
                # This replica should be taken off handle's replica set.
                transitioned_backend_tags.add(backend_tag)
                replica.stop()
                replicas.add(ReplicaState.STOPPING, replica)

            slow_start_replicas = []
            for replica in replicas.pop(
                    states=[ReplicaState.STARTING_OR_UPDATING]):
                ready, slow_start = replica.check_started()
                if ready:
                    # This replica should be now be added to handle's replica
                    # set.
                    replicas.add(ReplicaState.RUNNING, replica)
                    transitioned_backend_tags.add(backend_tag)
                else:
                    replicas.add(ReplicaState.STARTING_OR_UPDATING, replica)
                    if slow_start:
                        slow_start_replicas.append(replica)

            if (len(slow_start_replicas)
                    and time.time() - self._prev_startup_warnings[backend_tag]
                    > SLOW_STARTUP_WARNING_PERIOD_S):
                required, available = slow_start_replicas[
                    0].resource_requirements()
                logger.warning(
                    f"Backend '{backend_tag}' has {len(slow_start_replicas)} "
                    "replicas that have taken more than "
                    f"{SLOW_STARTUP_WARNING_S}s to start up. This may be "
                    "caused by waiting for the cluster to auto-scale or "
                    "because the constructor is slow. Resources required "
                    f"for each replica: {required}, resources available: "
                    f"{available}.")

                self._prev_startup_warnings[backend_tag] = time.time()

            for replica in replicas.pop(states=[ReplicaState.STOPPING]):
                if not replica.check_stopped():
                    replicas.add(ReplicaState.STOPPING, replica)

        if len(transitioned_backend_tags) > 0:
            self._checkpoint()
            [
                self._notify_replica_handles_changed(tag)
                for tag in transitioned_backend_tags
            ]



import argparse
from gym.spaces import Box, Discrete
import numpy as np

from ray.rllib.examples.models.custom_model_api import DuelingQModel, \
    TorchDuelingQModel, ContActionQModel, TorchContActionQModel
from ray.rllib.models.catalog import ModelCatalog, MODEL_DEFAULTS
from ray.rllib.utils.framework import try_import_tf, try_import_torch

tf1, tf, tfv = try_import_tf()
torch, _ = try_import_torch()

parser = argparse.ArgumentParser()
parser.add_argument(
    "--framework",
    choices=["tf", "tf2", "tfe", "torch"],
    default="tf",
    help="The DL framework specifier.")

if __name__ == "__main__":
    args = parser.parse_args()

    # Test API wrapper for dueling Q-head.

    obs_space = Box(-1.0, 1.0, (3, ))
    action_space = Discrete(3)

    # Run in eager mode for value checking and debugging.
    tf1.enable_eager_execution()

    # __sphinx_doc_model_construct_1_begin__
    my_dueling_model = ModelCatalog.get_model_v2(
        obs_space=obs_space,
        action_space=action_space,
        num_outputs=action_space.n,
        model_config=MODEL_DEFAULTS,
        framework=args.framework,
        # Providing the `model_interface` arg will make the factory
        # wrap the chosen default model with our new model API class
        # (DuelingQModel). This way, both `forward` and `get_q_values`
        # are available in the returned class.
        model_interface=DuelingQModel
        if args.framework != "torch" else TorchDuelingQModel,
        name="dueling_q_model",
    )
    # __sphinx_doc_model_construct_1_end__

    batch_size = 10
    input_ = np.array([obs_space.sample() for _ in range(batch_size)])
    # Note that for PyTorch, you will have to provide torch tensors here.
    if args.framework == "torch":
        input_ = torch.from_numpy(input_)

    input_dict = {
        "obs": input_,
        "is_training": False,
    }
    out, state_outs = my_dueling_model(input_dict=input_dict)
    assert out.shape == (10, 256)
    # Pass `out` into `get_q_values`
    q_values = my_dueling_model.get_q_values(out)
    assert q_values.shape == (10, action_space.n)

    # Test API wrapper for single value Q-head from obs/action input.

    obs_space = Box(-1.0, 1.0, (3, ))
    action_space = Box(-1.0, -1.0, (2, ))

    # __sphinx_doc_model_construct_2_begin__
    my_cont_action_q_model = ModelCatalog.get_model_v2(
        obs_space=obs_space,
        action_space=action_space,
        num_outputs=2,
        model_config=MODEL_DEFAULTS,
        framework=args.framework,
        # Providing the `model_interface` arg will make the factory
        # wrap the chosen default model with our new model API class
        # (DuelingQModel). This way, both `forward` and `get_q_values`
        # are available in the returned class.
        model_interface=ContActionQModel
        if args.framework != "torch" else TorchContActionQModel,
        name="cont_action_q_model",
    )
    # __sphinx_doc_model_construct_2_end__

    batch_size = 10
    input_ = np.array([obs_space.sample() for _ in range(batch_size)])

    # Note that for PyTorch, you will have to provide torch tensors here.
    if args.framework == "torch":
        input_ = torch.from_numpy(input_)

    input_dict = {
        "obs": input_,
        "is_training": False,
    }
    # Note that for PyTorch, you will have to provide torch tensors here.
    out, state_outs = my_cont_action_q_model(input_dict=input_dict)
    assert out.shape == (10, 256)
    # Pass `out` and an action into `my_cont_action_q_model`
    action = np.array([action_space.sample() for _ in range(batch_size)])
    if args.framework == "torch":
        action = torch.from_numpy(action)

    q_value = my_cont_action_q_model.get_single_q_value(out, action)
    assert q_value.shape == (10, 1)



from typing import Callable, List

from ray.experimental.data.impl.block import Block, BlockMetadata, ObjectRef, T
from ray.experimental.data.impl.block_list import BlockList


class LazyBlockList(BlockList[T]):
    def __init__(self, calls: Callable[[], ObjectRef[Block]],
                 metadata: List[BlockMetadata]):
        assert len(calls) == len(metadata), (calls, metadata)
        self._calls = calls
        self._blocks = [calls[0]()] if calls else []
        self._metadata = metadata

    def __len__(self):
        return len(self._calls)

    def __iter__(self):
        outer = self

        class Iter:
            def __init__(self):
                self._pos = -1

            def __iter__(self):
                return self

            def __next__(self):
                self._pos += 1
                if self._pos < len(outer._calls):
                    return outer._get_or_compute(self._pos)
                raise StopIteration

        return Iter()

    def _get_or_compute(self, i: int) -> ObjectRef[Block]:
        assert i < len(self._calls), i
        # Check if we need to compute more blocks.
        if i >= len(self._blocks):
            start = len(self._blocks)
            # Exponentially increase the number of blocks computed per batch.
            for c in self._calls[start:max(i + 1, start * 2)]:
                self._blocks.append(c())
        return self._blocks[i]



from ray.rllib.env.wrappers.model_vector_env import model_vector_env as mve
from ray.rllib.utils.deprecation import deprecation_warning

deprecation_warning(
    old="ray.rllib.env.model_vector_env.model_vector_env",
    new="ray.rllib.env.wrappers.model_vector_env.model_vector_env",
    error=False,
)

model_vector_env = mve



import os
import torch
import torch.nn as nn
import argparse

from filelock import FileLock
from torch.utils.data import DataLoader, Subset
from torchvision.datasets import CIFAR10
import torchvision.transforms as transforms

from tqdm import trange

import ray
from ray.util.sgd.torch import TorchTrainer, TrainingOperator
from ray.util.sgd.torch.resnet import ResNet18
from ray.util.sgd.utils import BATCH_SIZE, override


def initialization_hook():
    # Need this for avoiding a connection restart issue on AWS.
    os.environ["NCCL_SOCKET_IFNAME"] = "^docker0,lo"
    os.environ["NCCL_LL_THRESHOLD"] = "0"

    # set the below if needed
    # print("NCCL DEBUG SET")
    # os.environ["NCCL_DEBUG"] = "INFO"


class CifarTrainingOperator(TrainingOperator):
    @override(TrainingOperator)
    def setup(self, config):
        # Create model.
        model = ResNet18(config)

        # Create optimizer.
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=config.get("lr", 0.1),
            momentum=config.get("momentum", 0.9))

        # Load in training and validation data.
        transform_train = transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465),
                                 (0.2023, 0.1994, 0.2010)),
        ])  # meanstd transformation

        transform_test = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465),
                                 (0.2023, 0.1994, 0.2010)),
        ])
        with FileLock(".ray.lock"):
            train_dataset = CIFAR10(
                root="~/data",
                train=True,
                download=True,
                transform=transform_train)
            validation_dataset = CIFAR10(
                root="~/data",
                train=False,
                download=False,
                transform=transform_test)

        if config["test_mode"]:
            train_dataset = Subset(train_dataset, list(range(64)))
            validation_dataset = Subset(validation_dataset, list(range(64)))

        train_loader = DataLoader(
            train_dataset, batch_size=config[BATCH_SIZE], num_workers=2)
        validation_loader = DataLoader(
            validation_dataset, batch_size=config[BATCH_SIZE], num_workers=2)

        # Create scheduler.
        scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer, milestones=[150, 250, 350], gamma=0.1)

        # Create loss.
        criterion = nn.CrossEntropyLoss()

        # Register all components.
        self.model, self.optimizer, self.criterion, self.scheduler = \
            self.register(models=model, optimizers=optimizer,
                          criterion=criterion, schedulers=scheduler)
        self.register_data(
            train_loader=train_loader, validation_loader=validation_loader)


def train_cifar(test_mode=False,
                num_workers=1,
                use_gpu=False,
                num_epochs=5,
                fp16=False):
    trainer1 = TorchTrainer(
        training_operator_cls=CifarTrainingOperator,
        initialization_hook=initialization_hook,
        num_workers=num_workers,
        config={
            "lr": 0.1,
            "test_mode": test_mode,  # subset the data
            # this will be split across workers.
            BATCH_SIZE: 128 * num_workers
        },
        use_gpu=use_gpu,
        scheduler_step_freq="epoch",
        use_fp16=fp16,
        use_tqdm=False)
    pbar = trange(num_epochs, unit="epoch")
    for i in pbar:
        info = {"num_steps": 1} if test_mode else {}
        info["epoch_idx"] = i
        info["num_epochs"] = num_epochs
        # Increase `max_retries` to turn on fault tolerance.
        trainer1.train(max_retries=1, info=info)
        val_stats = trainer1.validate()
        pbar.set_postfix(dict(acc=val_stats["val_accuracy"]))

    print(trainer1.validate())
    trainer1.shutdown()
    print("success!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--address",
        required=False,
        type=str,
        help="the address to use for connecting to the Ray cluster")
    parser.add_argument(
        "--server-address",
        type=str,
        default=None,
        required=False,
        help="The address of server to connect to if using "
        "Ray Client.")
    parser.add_argument(
        "--num-workers",
        "-n",
        type=int,
        default=1,
        help="Sets number of workers for training.")
    parser.add_argument(
        "--num-epochs", type=int, default=5, help="Number of epochs to train.")
    parser.add_argument(
        "--use-gpu",
        action="store_true",
        default=False,
        help="Enables GPU training")
    parser.add_argument(
        "--fp16",
        action="store_true",
        default=False,
        help="Enables FP16 training. Requires `use-gpu`.")
    parser.add_argument(
        "--smoke-test",
        action="store_true",
        default=False,
        help="Finish quickly for testing.")
    parser.add_argument(
        "--tune", action="store_true", default=False, help="Tune training")

    args, _ = parser.parse_known_args()

    if args.server_address:
        ray.util.connect(args.server_address)
    else:
        num_cpus = 4 if args.smoke_test else None
        ray.init(address=args.address, num_cpus=num_cpus, log_to_driver=True)

    train_cifar(
        test_mode=args.smoke_test,
        num_workers=args.num_workers,
        use_gpu=args.use_gpu,
        num_epochs=args.num_epochs,
        fp16=args.fp16)



def override(cls):
    """Annotation for documenting method overrides.

    Args:
        cls (type): The superclass that provides the overridden method. If this
            cls does not actually have the method, an error is raised.
    """

    def check_override(method):
        if method.__name__ not in dir(cls):
            raise NameError("{} does not override any method of {}".format(
                method, cls))
        return method

    return check_override


def PublicAPI(obj):
    """Annotation for documenting public APIs.

    Public APIs are classes and methods exposed to end users of RLlib. You
    can expect these APIs to remain stable across RLlib releases.

    Subclasses that inherit from a ``@PublicAPI`` base class can be
    assumed part of the RLlib public API as well (e.g., all trainer classes
    are in public API because Trainer is ``@PublicAPI``).

    In addition, you can assume all trainer configurations are part of their
    public API as well.
    """

    return obj


def DeveloperAPI(obj):
    """Annotation for documenting developer APIs.

    Developer APIs are classes and methods explicitly exposed to developers
    for the purposes of building custom algorithms or advanced training
    strategies on top of RLlib internals. You can generally expect these APIs
    to be stable sans minor changes (but less stable than public APIs).

    Subclasses that inherit from a ``@DeveloperAPI`` base class can be
    assumed part of the RLlib developer API as well.
    """

    return obj



import warnings

import ray


# TODO(ekl) deprecate this after Ray 1.4
def get_node_ip_address(address="8.8.8.8:53"):
    warnings.warn(
        DeprecationWarning(
            "ray.services.get_node_ip_address has been moved to "
            "ray.util.get_node_ip_address. 'ray.services' will be "
            "removed after Ray 1.4."),
        stacklevel=2)
    return ray.util.get_node_ip_address(address)



import os
import sys
import unittest

import pytest
import ray
from ray import tune
from ray.job_config import JobConfig
from ray.rllib.agents import ppo
from ray.rllib.examples.env.stateless_cartpole import StatelessCartPole
from ray.rllib.utils.test_utils import check_learning_achieved
from ray.util.client.ray_client_helpers import ray_start_client_server


class TestRayClient(unittest.TestCase):
    def test_connection(self):
        with ray_start_client_server():
            assert ray.util.client.ray.is_connected()
        assert ray.util.client.ray.is_connected() is False

    def test_custom_train_fn(self):
        with ray_start_client_server():
            assert ray.util.client.ray.is_connected()

            config = {
                "lr": 0.01,
                # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
                "num_gpus": int(os.environ.get("RLLIB_NUM_GPUS", "0")),
                "num_workers": 0,
                "framework": "tf",
            }
            resources = ppo.PPOTrainer.default_resource_request(config)
            from ray.rllib.examples.custom_train_fn import my_train_fn
            tune.run(my_train_fn, resources_per_trial=resources, config=config)

    def test_cartpole_lstm(self):
        with ray_start_client_server():
            assert ray.util.client.ray.is_connected()

            config = dict(
                {
                    "num_sgd_iter": 5,
                    "model": {
                        "vf_share_layers": True,
                    },
                    "vf_loss_coeff": 0.0001,
                },
                **{
                    "env": StatelessCartPole,
                    # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
                    "num_gpus": int(os.environ.get("RLLIB_NUM_GPUS", "0")),
                    "model": {
                        "use_lstm": True,
                        "lstm_cell_size": 256,
                        "lstm_use_prev_action": None,
                        "lstm_use_prev_reward": None,
                    },
                    "framework": "tf",
                    # Run with tracing enabled for tfe/tf2?
                    "eager_tracing": None,
                })

            stop = {
                "training_iteration": 200,
                "timesteps_total": 100000,
                "episode_reward_mean": 150.0,
            }

            results = tune.run("PPO", config=config, stop=stop, verbose=2)
            check_learning_achieved(results, 150.0)

    def test_custom_experiment(self):
        def ray_connect_handler(job_config: JobConfig = None):
            ray.init(num_cpus=3)

        with ray_start_client_server(ray_connect_handler=ray_connect_handler):
            assert ray.util.client.ray.is_connected()

            config = ppo.DEFAULT_CONFIG.copy()
            config["train-iterations"] = 10
            config["env"] = "CartPole-v0"

            from ray.rllib.examples.custom_experiment import experiment
            tune.run(
                experiment,
                config=config,
                resources_per_trial=ppo.PPOTrainer.default_resource_request(
                    config))


if __name__ == "__main__":
    sys.exit(pytest.main(["-v", __file__]))



# ---------------
# IMPORTANT NOTE:
# ---------------
# A recent bug in openAI gym prevents RLlib's "record_env" option
# from recording videos properly. Instead, the produced mp4 files
# have a size of 1kb and are corrupted.
# A simple fix for this is described here:
# https://github.com/openai/gym/issues/1925

import argparse
import gym
import numpy as np
import ray
from gym.spaces import Box, Discrete

from ray import tune
from ray.rllib.env.multi_agent_env import make_multi_agent

parser = argparse.ArgumentParser()
parser.add_argument(
    "--framework",
    choices=["tf", "tf2", "tfe", "torch"],
    default="tf",
    help="The DL framework specifier.")
parser.add_argument("--multi-agent", action="store_true")
parser.add_argument("--stop-iters", type=int, default=10)
parser.add_argument("--stop-timesteps", type=int, default=10000)
parser.add_argument("--stop-reward", type=float, default=9.0)


class CustomRenderedEnv(gym.Env):
    """Example of a custom env, for which you can specify rendering behavior.
    """

    # Must specify, which render modes are supported by your custom env.
    # For RLlib to render your env via the "render_env" config key, only
    # mode="rgb_array" is needed. RLlib will automatically produce a simple
    # viewer for the returned RGB-images for mode="human", such that you don't
    # have to provide your own window+render handling.
    metadata = {
        "render.modes": ["rgb_array"],
    }

    def __init__(self, config):
        self.end_pos = config.get("corridor_length", 10)
        self.max_steps = config.get("max_steps", 100)
        self.cur_pos = 0
        self.steps = 0
        self.action_space = Discrete(2)
        self.observation_space = Box(0.0, 999.0, shape=(1, ), dtype=np.float32)

    def reset(self):
        self.cur_pos = 0.0
        self.steps = 0
        return [self.cur_pos]

    def step(self, action):
        self.steps += 1
        assert action in [0, 1], action
        if action == 0 and self.cur_pos > 0:
            self.cur_pos -= 1.0
        elif action == 1:
            self.cur_pos += 1.0
        done = self.cur_pos >= self.end_pos or \
            self.steps >= self.max_steps
        return [self.cur_pos], 10.0 if done else -0.1, done, {}

    def render(self, mode="rgb"):
        """Implements rendering logic for this env (given current state).

        You can either return an RGB image:
        np.array([height, width, 3], dtype=np.uint8) or take care of
        rendering in a window yourself here (return True then).
        For RLlib, though, only mode=rgb (returning an image) is needed,
        even when "render_env" is True in the RLlib config.

        Args:
            mode (str): One of "rgb", "human", or "ascii". See gym.Env for
                more information.

        Returns:
            Union[np.ndarray, bool]: An image to render or True (if rendering
                is handled entirely in here).
        """

        # Just generate a random image here for demonstration purposes.
        # Also see `gym/envs/classic_control/cartpole.py` for
        # an example on how to use a Viewer object.
        return np.random.randint(0, 256, size=(300, 400, 3), dtype=np.uint8)


MultiAgentCustomRenderedEnv = make_multi_agent(
    lambda config: CustomRenderedEnv(config))

if __name__ == "__main__":
    # Note: Recording and rendering in this example
    # should work for both local_mode=True|False.
    ray.init(num_cpus=4)
    args = parser.parse_args()

    # Example config causing
    config = {
        # Also try common gym envs like: "CartPole-v0" or "Pendulum-v0".
        "env": (MultiAgentCustomRenderedEnv
                if args.multi_agent else CustomRenderedEnv),
        "env_config": {
            "corridor_length": 10,
            "max_steps": 100,
        },
        # Evaluate once per training iteration.
        "evaluation_interval": 1,
        # Run evaluation on (at least) two episodes
        "evaluation_num_episodes": 2,
        # ... using one evaluation worker (setting this to 0 will cause
        # evaluation to run on the local evaluation worker, blocking
        # training until evaluation is done).
        "evaluation_num_workers": 1,
        # Special evaluation config. Keys specified here will override
        # the same keys in the main config, but only for evaluation.
        "evaluation_config": {
            # Store videos in this relative directory here inside
            # the default output dir (~/ray_results/...).
            # Alternatively, you can specify an absolute path.
            # Set to True for using the default output dir (~/ray_results/...).
            # Set to False for not recording anything.
            "record_env": "videos",
            # "record_env": "/Users/xyz/my_videos/",

            # Render the env while evaluating.
            # Note that this will always only render the 1st RolloutWorker's
            # env and only the 1st sub-env in a vectorized env.
            "render_env": True,
        },
        "num_workers": 1,
        # Use a vectorized env with 2 sub-envs.
        "num_envs_per_worker": 2,
        "framework": args.framework,
    }

    stop = {
        "training_iteration": args.stop_iters,
        "timesteps_total": args.stop_timesteps,
        "episode_reward_mean": args.stop_reward,
    }

    results = tune.run("PPO", config=config, stop=stop)



import functools
import inspect
from typing import List, Tuple, Callable, Optional

import ray

from ray import ObjectRef
from ray.experimental.workflow import serialization_context
from ray.experimental.workflow.common import (
    Workflow, StepID, WorkflowOutputType, WorkflowInputTuple)
from ray.experimental.workflow.step_executor import execute_workflow_step

StepInputTupleToResolve = Tuple[ObjectRef, List[ObjectRef], List[ObjectRef]]


class WorkflowStepFunction:
    """This class represents a workflow step."""

    def __init__(self, func: Callable):
        self._func = func
        self._func_signature = list(
            inspect.signature(func).parameters.values())

        # Override signature and docstring
        @functools.wraps(func)
        def _build_workflow(*args, **kwargs) -> Workflow:
            # validate if the input arguments match the signature of the
            # original function.
            reconstructed_signature = inspect.Signature(
                parameters=self._func_signature)
            try:
                reconstructed_signature.bind(*args, **kwargs)
            except TypeError as exc:  # capture a friendlier stacktrace
                raise TypeError(str(exc)) from None
            workflows: List[Workflow] = []
            object_refs: List[ObjectRef] = []
            with serialization_context.workflow_args_serialization_context(
                    workflows, object_refs):
                # NOTE: When calling 'ray.put', we trigger python object
                # serialization. Under our serialization context,
                # Workflows and ObjectRefs are separated from the arguments,
                # leaving a placeholder object with all other python objects.
                # Then we put the placeholder object to object store,
                # so it won't be mutated later. This guarantees correct
                # semantics. See "tests/test_variable_mutable.py" as
                # an example.
                input_placeholder: ObjectRef = ray.put((args, kwargs))
                if object_refs:
                    raise ValueError(
                        "There are ObjectRefs in workflow inputs. However "
                        "workflow currently does not support checkpointing "
                        "ObjectRefs.")
            return Workflow(self._func, self._run_step, input_placeholder,
                            workflows, object_refs)

        self.step = _build_workflow

    def _run_step(
            self,
            step_id: StepID,
            step_inputs: WorkflowInputTuple,
            outer_most_step_id: Optional[StepID] = None) -> WorkflowOutputType:
        return execute_workflow_step(self._func, step_id, step_inputs,
                                     outer_most_step_id)

    def __call__(self, *args, **kwargs):
        raise TypeError("Workflow steps cannot be called directly. Instead "
                        f"of running '{self.step.__name__}()', "
                        f"try '{self.step.__name__}.step()'.")



#!/usr/bin/env python

import numpy as np
import argparse
import json
import os
import random

import ray
from ray import tune
from ray.tune.schedulers import PopulationBasedTraining


def pbt_function(config, checkpoint_dir=None):
    """Toy PBT problem for benchmarking adaptive learning rate.

    The goal is to optimize this trainable's accuracy. The accuracy increases
    fastest at the optimal lr, which is a function of the current accuracy.

    The optimal lr schedule for this problem is the triangle wave as follows.
    Note that many lr schedules for real models also follow this shape:

     best lr
      ^
      |    /\
      |   /  \
      |  /    \
      | /      \
      ------------> accuracy

    In this problem, using PBT with a population of 2-4 is sufficient to
    roughly approximate this lr schedule. Higher population sizes will yield
    faster convergence. Training will not converge without PBT.
    """
    lr = config["lr"]
    accuracy = 0.0  # end = 1000
    start = 0
    if checkpoint_dir:
        with open(os.path.join(checkpoint_dir, "checkpoint")) as f:
            state = json.loads(f.read())
            accuracy = state["acc"]
            start = state["step"]

    midpoint = 100  # lr starts decreasing after acc > midpoint
    q_tolerance = 3  # penalize exceeding lr by more than this multiple
    noise_level = 2  # add gaussian noise to the acc increase
    # triangle wave:
    #  - start at 0.001 @ t=0,
    #  - peak at 0.01 @ t=midpoint,
    #  - end at 0.001 @ t=midpoint * 2,
    for step in range(start, 100):
        if accuracy < midpoint:
            optimal_lr = 0.01 * accuracy / midpoint
        else:
            optimal_lr = 0.01 - 0.01 * (accuracy - midpoint) / midpoint
        optimal_lr = min(0.01, max(0.001, optimal_lr))

        # compute accuracy increase
        q_err = max(lr, optimal_lr) / min(lr, optimal_lr)
        if q_err < q_tolerance:
            accuracy += (1.0 / q_err) * random.random()
        elif lr > optimal_lr:
            accuracy -= (q_err - q_tolerance) * random.random()
        accuracy += noise_level * np.random.normal()
        accuracy = max(0, accuracy)

        if step % 3 == 0:
            with tune.checkpoint_dir(step=step) as checkpoint_dir:
                path = os.path.join(checkpoint_dir, "checkpoint")
                with open(path, "w") as f:
                    f.write(json.dumps({"acc": accuracy, "step": start}))

        tune.report(
            mean_accuracy=accuracy,
            cur_lr=lr,
            optimal_lr=optimal_lr,  # for debugging
            q_err=q_err,  # for debugging
            done=accuracy > midpoint * 2  # this stops the training process
        )


def run_tune_pbt():
    pbt = PopulationBasedTraining(
        time_attr="training_iteration",
        perturbation_interval=4,
        hyperparam_mutations={
            # distribution for resampling
            "lr": lambda: random.uniform(0.0001, 0.02),
            # allow perturbations within this set of categorical values
            "some_other_factor": [1, 2],
        })

    analysis = tune.run(
        pbt_function,
        name="pbt_test",
        scheduler=pbt,
        verbose=False,
        metric="mean_accuracy",
        mode="max",
        stop={
            "training_iteration": 30,
        },
        num_samples=8,
        fail_fast=True,
        config={
            "lr": 0.0001,
            # note: this parameter is perturbed but has no effect on
            # the model training in this example
            "some_other_factor": 1,
        })

    print("Best hyperparameters found were: ", analysis.best_config)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--smoke-test", action="store_true", help="Finish quickly for testing")
    parser.add_argument(
        "--server-address",
        type=str,
        default=None,
        required=False,
        help="The address of server to connect to if using "
        "Ray Client.")
    args, _ = parser.parse_known_args()
    if args.smoke_test:
        ray.init(num_cpus=2)  # force pausing to happen for test
    else:
        if args.server_address is not None:
            ray.util.connect(args.server_address)
        else:
            ray.init()

    run_tune_pbt()
